import uuid
import sys
from importlib import reload

from ...libs.aliyunsdkdysmsapi.request.v20170525 import SendSmsRequest
from ...libs.aliyunsdkdysmsapi.request.v20170525 import QuerySendDetailsRequest
from aliyunsdkcore.client import AcsClient
from aliyunsdkcore.profile import region_provider
from aliyunsdkcore.http import method_type as MT
from aliyunsdkcore.http import format_type as FT

from . import const

"""
短信业务调用接口示例，版本号：v20170525

Created on 2017-06-12

"""
try:
    reload(sys)
    # sys.setdefaultencoding('utf8')
except NameError:
    pass
except Exception as err:
    raise err

# 注意：不要更改
REGION = "cn-hangzhou"
PRODUCT_NAME = "Dysmsapi"
DOMAIN = "dysmsapi.aliyuncs.com"

acs_client = AcsClient(const.ACCESS_KEY_ID, const.ACCESS_KEY_SECRET, REGION)
region_provider.add_endpoint(PRODUCT_NAME, REGION, DOMAIN)


def send_sms(business_id, phone_numbers, sign_name, template_code, template_param=None):
    smsRequest = SendSmsRequest.SendSmsRequest()
    # 申请的短信模板编码,必填
    smsRequest.set_TemplateCode(template_code)

    # 短信模板变量参数
    if template_param is not None:
        smsRequest.set_TemplateParam(template_param)

    # 设置业务请求流水号，必填。
    smsRequest.set_OutId(business_id)

    # 短信签名
    smsRequest.set_SignName(sign_name)

    # 数据提交方式
    # smsRequest.set_method(MT.POST)

    # 数据提交格式
    # smsRequest.set_accept_format(FT.JSON)

    # 短信发送的号码列表，必填。
    smsRequest.set_PhoneNumbers(phone_numbers)

    # 调用短信发送接口，返回json
    smsResponse = acs_client.do_action_with_exception(smsRequest)

    # TODO 业务处理

    return smsResponse


if __name__ == '__main__':
    __business_id = uuid.uuid1()
    # print(__business_id)
    params = "{\"code\":\"12345\",\"product\":\"云通信\"}"
    # params = u'{"name":"wqb","code":"12345678","address":"bz","phone":"13000000000"}'
    print(send_sms(__business_id, "13000000000", "云通信测试", "SMS_5250008", params))



import uuid
from datetime import datetime

from django.core.exceptions import ValidationError
from django.db import models

from scaffold.exceptions.exceptions import AppError


# def patch_methods(model_class):
#     def do_patch(cls):
#         for k in cls.__dict__:
#             obj = getattr(cls, k)
#             if not k.startswith('_') and callable(obj):
#                 setattr(model_class, k, obj)
#
#     return do_patch

class UUIDModel(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    class Meta:
        abstract = True


class SortableModel(models.Model):
    """ 可排序模型
    """
    sorting = models.BigIntegerField(
        verbose_name='排序',
        default=0,
        help_text='用于系统进行排序的参数，可以给用户设定或者作为计算列存储组合权重',
        db_index=True,
    )

    class Meta:
        abstract = True
        ordering = ['-sorting']


class StickModel(models.Model):
    """ 可置顶模型
    """
    is_sticky = models.BooleanField(
        verbose_name='是否置顶',
        default=False,
        db_index=True,
    )

    class Meta:
        abstract = True
        ordering = ['-is_sticky']


class ActiveModel(models.Model):
    """ 可以切换可用/不可用的模型
    """
    is_active = models.BooleanField(
        verbose_name='是否可用',
        default=True,
        db_index=True,
    )

    class Meta:
        abstract = True


class DatedModel(models.Model):
    """ 记录了创建时间和修改时间的模型
    """
    date_created = models.DateTimeField(
        verbose_name='创建时间',
        auto_now_add=True,
        db_index=True,
    )

    date_updated = models.DateTimeField(
        verbose_name='修改时间',
        auto_now=True,
        db_index=True,
    )

    class Meta:
        abstract = True


class NamedModel(models.Model):
    """ 有名称的模型
    """

    name = models.CharField(
        verbose_name='名称',
        max_length=255,
        blank=True,
        default='',
    )

    class Meta:
        abstract = True

    def __str__(self):
        return self.name or '[{}]'.format(self.pk)


class ContentModel(models.Model):
    """ 有内容的模型
    """

    content = models.TextField(
        verbose_name='内容',
        blank=True,
        default='',
    )

    excerpt = models.CharField(
        verbose_name='摘要',
        max_length=255,
        blank=True,
        default='',
    )

    class Meta:
        abstract = True


class HierarchicalModel(models.Model):
    """ 层次模型，具备 parent 和 children 属性
    """
    parent = models.ForeignKey(
        verbose_name='上级',
        to='self',
        related_name='children',
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )

    class Meta:
        abstract = True

    def clean(self):
        # 环路检测
        p = self.parent
        while p is not None:
            if p.pk == self.pk:
                raise ValidationError('级联结构不能出现循环引用')
            p = p.parent

    @property
    def parent_name(self):
        return self.parent and getattr(self.parent, 'name', None)


class NullableUserOwnedModel(models.Model):
    """ 由用户拥有的模型类
    包含作者字段
    """

    author = models.ForeignKey(
        verbose_name='作者',
        to='auth.User',
        related_name='%(class)ss_owned',
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )

    class Meta:
        abstract = True


class UserOwnedModel(models.Model):
    """ 由用户拥有的模型类
    包含作者字段，要求非空
    """

    author = models.ForeignKey(
        verbose_name='作者',
        to='auth.User',
        related_name='%(class)ss_owned',
        on_delete=models.CASCADE,
    )

    class Meta:
        abstract = True


class EntityModel(NamedModel,
                  SortableModel,
                  StickModel,
                  DatedModel):
    """ 实体类模型
    """

    class Meta:
        abstract = True
        ordering = ['-date_created', '-is_sticky', '-sorting']

    def __str__(self):
        return self.name or str(self.pk)


class AbstractValidationModel(models.Model):
    """ 抽象验证类
    1. 提交一次验证的时候，必须没有非 EXPIRED 的验证信息；
    2. 提交验证之后，创建一条新的 PersonalValidationInfo 信息；
    3. 新提交的验证，状态为 PENDING，记录 date_submitted；
    4. 管理员权限可以进行审批，或者驳回，改变状态并记录 date_response；
    5. 任何阶段，用户可以取消掉现有的验证信息，变成 EXPIRED 并记录时间；
    6. 取消掉唯一一条活动的验证信息之后，可以提交新的验证信息；
    """

    STATUS_DRAFT = 'DRAFT'
    STATUS_PENDING = 'PENDING'
    STATUS_REJECTED = 'REJECTED'
    STATUS_SUCCESS = 'SUCCESS'
    STATUS_EXPIRED = 'EXPIRED'
    STATUS_CHOICES = (
        (STATUS_DRAFT, '草稿'),
        (STATUS_PENDING, '等待审批'),
        (STATUS_REJECTED, '驳回'),
        (STATUS_SUCCESS, '成功'),
        (STATUS_EXPIRED, '已失效'),
    )

    status = models.CharField(
        verbose_name='验证状态',
        max_length=20,
        choices=STATUS_CHOICES,
        default=STATUS_DRAFT,
    )

    date_submitted = models.DateTimeField(
        verbose_name='提交时间',
        blank=True,
        null=True,
    )

    date_response = models.DateTimeField(
        verbose_name='审批时间',
        blank=True,
        null=True,
    )

    date_expired = models.DateTimeField(
        verbose_name='失效时间',
        blank=True,
        null=True,
    )

    remark = models.CharField(
        verbose_name='审核不通过原因',
        max_length=255,
        blank=True,
        default='',
    )

    class Meta:
        abstract = True

    def approve(self, *args, **kwargs):
        if self.status not in (self.STATUS_PENDING, self.STATUS_REJECTED):
            raise AppError('ERR091', '审批对象的状态必须为等待审批或者驳回')
        self.status = self.STATUS_SUCCESS
        self.date_response = datetime.now()
        self.save()

    def reject(self, reason, *args, **kwargs):
        if self.status not in (self.STATUS_PENDING,):
            raise AppError('ERR092', '审批对象的状态必须为等待审批')
        if not reason:
            raise AppError('ERR093', '请填写驳回理由')
        self.status = self.STATUS_REJECTED
        self.date_response = datetime.now()
        self.remark = reason
        self.save()


class AbstractTransactionModel(models.Model):
    debit = models.ForeignKey(
        verbose_name='借方用户',
        to='auth.User',
        related_name='%(class)ss_debit',
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        help_text='即余额增加的账户，默认情况用户作为账户，'
                  '如需定义其他模型作为账号，派生时覆写此字段',
    )

    credit = models.ForeignKey(
        verbose_name='贷方用户',
        to='auth.User',
        related_name='%(class)ss_credit',
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        help_text='即余额减少的账户，默认情况用户作为账户，'
                  '如需定义其他模型作为账号，派生时覆写此字段',
    )

    amount = models.DecimalField(
        verbose_name='金额',
        max_digits=18,
        decimal_places=2,
    )

    remark = models.CharField(
        verbose_name='备注',
        blank=True,
        default='',
        max_length=255,
    )

    class Meta:
        abstract = True



from django.core.exceptions import ValidationError


class AppError(ValidationError):
    """ 应用错误
    可以自动映射到 HTTP 错误响应输出
    """
    data = None
    http_status = None
    silent = False
    debug = False

    def __init__(self, code, message, *, data=None, http_status=400, silent=False, debug=False):
        super().__init__(message, code)
        self.data = data
        self.http_status = http_status
        self.silent = silent
        self.debug = debug

    def set_data(self, data):
        self.data = data
        return self

    def set_silent(self, silent):
        self.silent = silent
        return self

    def set_message(self, message):
        self.message = message
        return self

    def set_debug(self, debug):
        self.debug = debug
        return self

    def set_status_code(self, status_code):
        self.http_status = status_code
        return self



""" Extended rest_framework authentication classes """
from rest_framework.authentication import SessionAuthentication


class CsrfExemptSessionAuthentication(SessionAuthentication):
    """ Replacing the Csrf Authentication """

    def enforce_csrf(self, request):
        return  # To not perform the csrf check previously happening



""" rewrite the DRF filters module """
from __future__ import unicode_literals

import operator
import re
import sys
from functools import reduce

from django.conf import settings
from django.db import models
from django.db.models import Q
from django.db.models.constants import LOOKUP_SEP
from django.template import loader
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from rest_framework.compat import distinct
from rest_framework.filters import BaseFilterBackend
from rest_framework.settings import api_settings

try:
    import coreapi
except ImportError:
    coreapi = None

try:
    import coreschema
except ImportError:
    coreschema = None


class DeepFilterBackend:
    """ 深入查询参数过滤器
    在 Django Rest Framework 中，支持以 queryset 的直接参数查询进行列表集的筛选
    在 DEFAULT_FILTER_BACKENDS 或者 ViewSet 的 filter_backends 中使用这个 DeepFilterBackend：

    'DEFAULT_FILTER_BACKENDS': (
        // ...
        'django_base.base_utils.filters.DeepFilterBackend',
    ),

    然后，在所有的 ListModelMixin 中，所有带有双下划线 `__` 的 querystring 参数，
    都会尝试在 queryset 中进行筛选。

    例如：

    ?user__username__startswith=admin_

    特殊处理：

    * 所有的 value 值如果是 True/False/None，会被转换成对应的 python 值
    * 如果查询参数前面带有 !，例如 `!name__startswith=a`，将使用 exclude 排除筛选条件而不是 filter

    为了安全起见，所有被捕获的查询条件参数，必须显式注册在对应的 View Class 的
    'allowed_deep_params' 列表里面，否则条件会被忽略并且获得一条警告。

    例如是的上述的两个条件生效：

    class UserViewSet(viewsets.ModelViewSet):
        # ...
        allowed_deep_params = ['user__username__startswith', 'name__starts__with']

    TODO: 考虑支持全部参数开放或者正则表达式过滤参数或者用户级别的权限控制
    TODO: 考虑添加静默选项以防止生产环境过多输出警告参数
    """

    request = None
    allowed_deep_params = []

    @staticmethod
    def never():
        """ todo """
        return Q(pk=None)

    @staticmethod
    def always():
        """ todo """
        return ~Q(pk=None)

    @staticmethod
    def get_setting_value(key, default):
        """ todo """
        if not hasattr(settings, key):
            return default
        return getattr(settings, key)

    def malformed_query(self):
        """ todo """
        # 配置 ALLOW_MALFORMED_QUERY = False 以实现非授权条件断路（返回空集）
        return self.always() if self.get_setting_value('ALLOW_MALFORMED_QUERY', True) \
            else self.never()

    def filter_queryset(self, request, queryset, view):
        """ todo """
        self.request = request
        self.allowed_deep_params = getattr(view, 'allowed_deep_params', ())

        # 注意，这个只能在 list 方法中生效，其他方法需要在 ViewSet 中加白名单
        if view.action != 'list' and \
                view.action not in getattr(view, 'allowed_filtering_actions', []):
            return queryset

        for key, val in request.query_params.items():
            if '__' in key or val.startswith('$F__'):
                queryset = queryset.filter(self.get_single_condition_query(key, val))
            if key.startswith('_complex_query'):
                queryset = queryset.filter(self.parse_complex_query(val))

        if getattr(settings, 'REST_DEEP_DEFAULT_DISTINCT', None) \
                or request.query_params.get('~DISTINCT'):
            queryset = queryset.distinct()

        # 最后要加 distinct 去重
        return queryset

    def get_single_condition_query(self, key, val):
        """ todo """
        # 不满足的条件设置为条件短路
        if not re.match(r'^!*[A-Za-z0-9_]+$', key):
            print('!!!! Unsupported query phase: {}={}'.format(key, val), file=sys.stderr)
            return self.malformed_query()
        if key[0] == '!':
            return ~self.get_single_condition_query(key[1:], val)
        # 管理员登录可以豁免，否则所有的级联搜索必须显式放行
        user = getattr(self.request, 'user')
        if not self.get_setting_value('ALLOW_ALL_DEEP_PARAMS', False) \
                and key not in self.allowed_deep_params \
                and (not user or not user.is_superuser):
            print(
                '!!!! Deep filter param not registered: ' + key + '\n' +
                'The param is skipped, to make it work, '
                'add the params key name to `allowed_deep_params` list '
                'in the View class.\n!!!!', file=sys.stderr)
            return self.malformed_query()

        # 特殊字符串值映射
        value_mapper = {'False': False, 'True': True, 'None': None}
        if val.startswith('$F__'):
            # $F__xxxx 语义映射
            val = models.F(val[4:])
        elif val in value_mapper:
            val = value_mapper[val]
        # 如果是 id 列表类的入参，按照逗号进行分割
        if key.endswith('__in') and re.match(r'^(?:\d+,)*\d+$', val):
            val = map(int, val.split(','))
        elif key.endswith('__in') and re.match(r'^(?:[\d\w\-_]+,)*[\d\w\-_]+$', val):
            val = val.split(',')
        # 返回展开的条件
        return Q(**{key: val})

    def parse_complex_query(self, query):
        """ todo """
        if '||' in query:
            query_set = self.never()
            for part in query.split('||'):
                query_set |= self.parse_complex_query(part)
            return query_set
        if '&&' in query:
            query_set = self.always()
            for part in query.split('&&'):
                query_set &= self.parse_complex_query(part)
            return query_set
        tup = query.split('=')
        if len(tup) != 2:
            print('!!!! Unsupported query phase: {}'.format(query), file=sys.stderr)
            return self.malformed_query()
        return self.get_single_condition_query(*tup)  # tup=(key,val)

    def get_schema_fields(self, view):
        """ todo """
        # This is not compatible with widgets where the query param differs from the
        # filter's attribute name. Notably, this includes `MultiWidget`, where query
        # params will be of the format `<name>_0`, `<name>_1`, etc...
        assert coreapi is not None, 'coreapi must be installed to use `get_schema_fields()`'
        assert coreschema is not None, 'coreschema must be installed to use `get_schema_fields()`'

        return [
            coreapi.Field(
                name=field,
                required=False,
                location='query',
                # schema=self.get_coreschema_field(field)
                schema=coreschema.String(description='* 注册的 DeepFilter 查询参数')
            ) for field in getattr(view, 'allowed_deep_params', [])
        ]


class SearchFilter(BaseFilterBackend):
    """ todo """
    # The URL query parameter used for the search.
    search_param = api_settings.SEARCH_PARAM
    template = 'rest_framework/filters/search.html'
    lookup_prefixes = {
        '^': 'istartswith',
        '=': 'iexact',
        '@': 'search',
        '$': 'iregex',
    }
    search_title = _('Search')
    search_description = _('A search term.')

    def get_search_terms(self, request):
        """
        Search terms are set by a ?search=... query parameter,
        and may be comma and/or whitespace delimited.
        """
        params = request.query_params.get(self.search_param, '')
        return params.replace(',', ' ').split()

    def construct_search(self, field_name):
        """ todo """
        lookup = self.lookup_prefixes.get(field_name[0])
        if lookup:
            field_name = field_name[1:]
        else:
            lookup = 'icontains'
        return LOOKUP_SEP.join([field_name, lookup])

    def must_call_distinct(self, queryset, search_fields):
        """
        Return True if 'distinct()' should be used to query the given lookups.
        """
        for search_field in search_fields:
            opts = queryset.model._meta
            if search_field[0] in self.lookup_prefixes:
                search_field = search_field[1:]
            parts = search_field.split(LOOKUP_SEP)
            for part in parts:
                field = opts.get_field(part)
                if hasattr(field, 'get_path_info'):
                    # This field is a relation, update opts to follow the relation
                    path_info = field.get_path_info()
                    opts = path_info[-1].to_opts
                    if any(path.m2m for path in path_info):
                        # This field is a m2m relation so we know we need to call distinct
                        return True
        return False

    def filter_queryset(self, request, queryset, view):
        """ todo """
        search_fields = getattr(view, 'search_fields', None)
        search_terms = self.get_search_terms(request)

        if not search_fields or not search_terms:
            return queryset

        orm_lookups = [
            self.construct_search(str(search_field))
            for search_field in search_fields
        ]

        base = queryset
        conditions = []
        for search_term in search_terms:
            queries = [
                models.Q(**{orm_lookup: search_term})
                for orm_lookup in orm_lookups
            ]
            conditions.append(reduce(operator.or_, queries))
        queryset = queryset.filter(reduce(operator.and_, conditions))

        if self.must_call_distinct(queryset, search_fields):
            # Filtering against a many-to-many field requires us to
            # call queryset.distinct() in order to avoid duplicate items
            # in the resulting queryset.
            # We try to avoid this if possible, for performance reasons.
            queryset = distinct(queryset, base)
        return queryset

    def to_html(self, request, queryset, view):
        """ todo """
        if not getattr(view, 'search_fields', None):
            return ''

        term = self.get_search_terms(request)
        term = term[0] if term else ''
        context = {
            'param': self.search_param,
            'term': term
        }
        template = loader.get_template(self.template)
        return template.render(context)

    def get_schema_fields(self, view):
        """ todo """
        assert coreapi is not None, 'coreapi must be installed to use `get_schema_fields()`'
        assert coreschema is not None, 'coreschema must be installed to use `get_schema_fields()`'
        return [
            coreapi.Field(
                name=self.search_param,
                required=False,
                location='query',
                schema=coreschema.String(
                    title=force_text(self.search_title),
                    description=force_text(self.search_description)
                )
            )
        ]


class OrderingFilter(BaseFilterBackend):
    """
    改编自 rest_framework.filters.OrderingFilter，由于过于严格的过滤，导致不能支持级联字段的排序，例如
    ordering=-related_item__name 的过滤，在这里将过滤的逻辑剔除，以使得功能增强。
    """
    # The URL query parameter used for the ordering.
    ordering_param = api_settings.ORDERING_PARAM
    ordering_fields = None
    ordering_title = _('Ordering')
    ordering_description = _('Which field to use when ordering the results.')
    template = 'rest_framework/filters/ordering.html'

    def get_ordering(self, request, queryset, view):
        """
        Ordering is set by a comma delimited ?ordering=... query parameter.

        The `ordering` query parameter can be overridden by setting
        the `ordering_param` value on the OrderingFilter or by
        specifying an `ORDERING_PARAM` value in the API settings.
        """
        params = request.query_params.get(self.ordering_param)
        if params:
            fields = [param.strip() for param in params.split(',')]
            return fields
            # ordering = self.remove_invalid_fields(queryset, fields, view, request)
            # if ordering:
            #     return ordering

        # No ordering was included, or all the ordering fields were invalid
        return self.get_default_ordering(view)

    def get_default_ordering(self, view):
        """ todo """
        ordering = getattr(view, 'ordering', None)
        if isinstance(ordering, str):
            return (ordering,)
        return ordering

    # def get_default_valid_fields(self, queryset, view, context={}):
    #     # If `ordering_fields` is not specified, then we determine a default
    #     # based on the serializer class, if one exists on the view.
    #     if hasattr(view, 'get_serializer_class'):
    #         try:
    #             serializer_class = view.get_serializer_class()
    #         except AssertionError:
    #             # Raised by the default implementation if
    #             # no serializer_class was found
    #             serializer_class = None
    #     else:
    #         serializer_class = getattr(view, 'serializer_class', None)
    #
    #     if serializer_class is None:
    #         msg = (
    #             "Cannot use %s on a view which does not have either a "
    #             "'serializer_class', an overriding 'get_serializer_class' "
    #             "or 'ordering_fields' attribute."
    #         )
    #         raise ImproperlyConfigured(msg % self.__class__.__name__)
    #
    #     return [
    #         (field.source.replace('.', '__') or field_name, field.label)
    #         for field_name, field in serializer_class(context=context).fields.items()
    #         if not getattr(field, 'write_only', False) and not field.source == '*'
    #     ]
    #
    # def get_valid_fields(self, queryset, view, context={}):
    #     valid_fields = getattr(view, 'ordering_fields', self.ordering_fields)
    #
    #     if valid_fields is None:
    #         # Default to allowing filtering on serializer fields
    #         return self.get_default_valid_fields(queryset, view, context)
    #
    #     elif valid_fields == '__all__':
    #         # View explicitly allows filtering on any model field
    #         valid_fields = [
    #             (field.name, field.verbose_name) for field in queryset.model._meta.fields
    #         ]
    #         valid_fields += [
    #             (key, key.title().split('__'))
    #             for key in queryset.query.annotations
    #         ]
    #     else:
    #         valid_fields = [
    #             (item, item) if isinstance(item, six.string_types) else item
    #             for item in valid_fields
    #         ]
    #
    #     return valid_fields

    # def remove_invalid_fields(self, queryset, fields, view, request):
    #     valid_fields = [item[0] for item in self.get_valid_fields(queryset, view,
    #  {'request': request})]
    #     return [term for term in fields if term.lstrip('-') in
    # valid_fields and ORDER_PATTERN.match(term)]

    def filter_queryset(self, request, queryset, view):
        """ todo """
        ordering = self.get_ordering(request, queryset, view)

        if ordering:
            return queryset.order_by(*ordering)

        return queryset

    def get_template_context(self, request, queryset, view):
        """ todo """
        current = self.get_ordering(request, queryset, view)
        current = None if not current else current[0]
        options = []
        context = {
            'request': request,
            'current': current,
            'param': self.ordering_param,
        }
        # for key, label in self.get_valid_fields(queryset, view, context):
        #     options.append((key, '%s - %s' % (label, _('ascending'))))
        #     options.append(('-' + key, '%s - %s' % (label, _('descending'))))
        context['options'] = options
        return context

    def to_html(self, request, queryset, view):
        """ todo """
        template = loader.get_template(self.template)
        context = self.get_template_context(request, queryset, view)
        return template.render(context)

    def get_schema_fields(self, view):
        """ todo """
        assert coreapi is not None, 'coreapi must be installed to use `get_schema_fields()`'
        assert coreschema is not None, 'coreschema must be installed to use `get_schema_fields()`'
        return [
            coreapi.Field(
                name=self.ordering_param,
                required=False,
                location='query',
                schema=coreschema.String(
                    title=force_text(self.ordering_title),
                    description=force_text(self.ordering_description)
                )
            )
        ]



from rest_framework import serializers
from ..media import serializers as media_serializer
from . import models as m


class OptionSerializer(serializers.ModelSerializer):
    class Meta:
        model = m.Option
        fields = '__all__'


class UserOptionSerializer(serializers.ModelSerializer):
    class Meta:
        model = m.UserOption
        fields = '__all__'


class VersionSerializer(serializers.ModelSerializer):
    attachment_item = media_serializer.AttachmentSerializer(source='attachment', read_only=True)

    class Meta:
        model = m.Version
        fields = '__all__'



# Generated by Django 3.0.8 on 2020-08-11 11:49

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('media', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Option',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=128, unique=True, verbose_name='选项关键字')),
                ('value', models.TextField(blank=True, default='', verbose_name='选项值')),
            ],
            options={
                'verbose_name': '系统选项',
                'verbose_name_plural': '系统选项',
                'db_table': 'base_config_option',
            },
        ),
        migrations.CreateModel(
            name='Version',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', models.CharField(max_length=20, verbose_name='版本号')),
                ('alias', models.CharField(blank=True, default='', max_length=100, verbose_name='版本别名')),
                ('platform', models.CharField(blank=True, default='', max_length=20, verbose_name='平台')),
                ('date_created', models.DateTimeField(auto_now_add=True, verbose_name='创建时间')),
                ('date_updated', models.DateTimeField(auto_now=True, verbose_name='更新时间')),
                ('is_active', models.BooleanField(default=True, verbose_name='是否可用')),
                ('is_master', models.BooleanField(default=True, help_text='建议实践：如果是主要版本，所有低版本的客户端必须要升级到此版本才能使用，否则为建议版本，只作提醒，不强制升级。', verbose_name='是否主要版本')),
                ('description', models.TextField(blank=True, default='', verbose_name='版本描述')),
                ('link', models.CharField(blank=True, default='', max_length=255, verbose_name='下载链接')),
                ('attachment', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='version', to='media.Attachment', verbose_name='安装包附件')),
            ],
            options={
                'verbose_name': '版本',
                'verbose_name_plural': '版本',
                'db_table': 'base_version',
            },
        ),
        migrations.CreateModel(
            name='UserOption',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=128, verbose_name='选项名')),
                ('value', models.TextField(blank=True, default='', verbose_name='选项值')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='options', to=settings.AUTH_USER_MODEL, verbose_name='用户')),
            ],
            options={
                'verbose_name': '用户选项',
                'verbose_name_plural': '用户选项',
                'db_table': 'base_config_user_option',
                'unique_together': {('user', 'key')},
            },
        ),
    ]



from django.utils.deprecation import MiddlewareMixin

from .exceptions import AppError
from ..utils.http import response_fail


class AppErrorMiddleware(MiddlewareMixin):
    @staticmethod
    def process_exception(request, exception):
        if type(exception) == AppError:
            import traceback
            from sys import stderr
            from django.conf import settings
            if exception.debug or settings.API_DEBUG:
                print(traceback.format_exc(), file=stderr)
            return response_fail(
                exception.message,
                exception.code,
                status=exception.http_status,
                data=exception.data,
                silent=exception.silent,
            )




from rest_framework import serializers


class DynamicFieldsModelSerializer(serializers.ModelSerializer):
    """
    支持动态指定字段的序列化器，传参fields，序列化和反序列化都支持
    参考：https://blog.csdn.net/hj009zzh/article/details/108850527
    """
    Meta: type

    def __init__(self, *args, **kwargs):
        """支持字段动态生成的序列化器，从默认的Meta.fields中过滤，无关字段不查不序列化"""
        fields = kwargs.pop('fields', None)
        super().__init__(*args, **kwargs)
        if fields is not None:
            allow = set(fields)
            existing = set(self.fields)
            for f in existing - allow:
                self.fields.pop(f)



from django.conf import settings


class AppRouter:
    db_map = getattr(settings, 'DATABASE_ROUTE_MAP', {})

    def db_for_read(self, model, **hints):
        """
        Attempts to read auth and contenttypes models go to auth_db.
        """
        return self.db_map.get(model._meta.app_label, 'default')

    def db_for_write(self, model, **hints):
        """
        Attempts to write auth and contenttypes models go to auth_db.
        """
        return self.db_map.get(model._meta.app_label, 'default')

    def allow_relation(self, obj1, obj2, **hints):
        """
        Allow relations if a model in the auth or contenttypes apps is
        involved.
        """
        return self.db_map.get(obj1._meta.app_label, 'default') == \
               self.db_map.get(obj2._meta.app_label, 'default') == 'default'

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        """
        Make sure the auth and contenttypes apps only appear in the
        'auth_db' database.
        """
        return self.db_map.get(app_label, 'default') == db == 'default'



import json
import re
import random
from functools import wraps
from time import time

from django.conf import settings
from django.core.exceptions import ValidationError


def sanitize_mobile(mobile):
    """ 过滤手机号码
    如果手机号码格式符合要求，直接返回
    否则抛出 ValidationError
    :param mobile:
    :return:
    """
    mobile = mobile.strip()

    from ..utils import is_valid_mobile
    assert is_valid_mobile(mobile), '手机号码格式不正确'
    return mobile


def validate_mobile_vcode(request, action, mobile, vcode):
    """ 要求 post 中提供 Session 中产生的验证码
    :param request: 请求对象
    :param action: 验证的 action 动作（不同的 action 独享一套 vcode）
    :param mobile: POST 提交过来的手机号码
    :param vcode: POST 提交过来的验证码
    :return: 如果成功返回 True，失败返回 False
    """

    from django_base.base_utils.app_error.exceptions import AppError
    _mobile, _vcode = get_vcode_info(request, action)

    if not vcode or vcode != _vcode:
        raise AppError(50001, '验证码不正确', debug=False)

    if not mobile or mobile != _mobile:
        raise AppError(50002, '手机号码不一致', debug=False)

    # 验证成功马上擦除
    clear_vcode_info(request)

    return True


def clear_vcode_info(request, action=''):
    prefix = ('mobile_vcode_' + action).strip('_')
    for key in [prefix, prefix + '_number', prefix + '_time']:
        if key in request.session:
            del request.session[key]
    request.session.modified = True


def set_vcode_info(request, mobile, vcode, action=''):
    prefix = ('mobile_vcode_' + action).strip('_')
    request.session[prefix + '_number'] = mobile
    request.session[prefix] = vcode
    request.session[prefix + '_time'] = int(time())
    request.session.modified = True


def get_vcode_info(request, action=''):
    """ 获取 Session 中存储的 验证码信息
    :param request:
    :param action:
    :return:
    """
    prefix = ('mobile_vcode_' + action).strip('_')
    # 上次请求验证码的时间
    last_sms_request_time = int(request.session.get(prefix + '_time', 0))
    # 验证码是否到期
    if time() > last_sms_request_time + settings.SMS_EXPIRE_INTERVAL:
        return None, None
    return request.session.get(prefix + '_number'), request.session.get(prefix)


def request_mobile_vcode(request, mobile, action=''):
    """
    为传入的 request 上下文（Session）产生一个手机验证码，
    发送给指定的手机号码，并且记录在 session.vcode 中。
    :param request:
    :param mobile:
    :return:
    """
    mobile = sanitize_mobile(mobile)

    prefix = 'mobile_vcode_' + action

    # 上次请求验证码的时间
    last_sms_request_time = int(request.session.get(prefix + '_time', 0))

    # 一分钟内不允许重发
    if time() < last_sms_request_time + settings.SMS_SEND_INTERVAL:
        from django_base.base_utils.app_error.exceptions import AppError
        raise AppError(
            99999,
            '您的操作过于频繁，请在 %d 秒后再试。' % (
                    last_sms_request_time + settings.SMS_SEND_INTERVAL - time()),
            debug=False
        )

    vcode = '%06d' % (random.randint(0, 1000000))

    # 如果开启了调试选项，不真正发送短信
    if not settings.SMS_DEBUG:
        sms_send(
            mobile,
            settings.SMS_TEMPLATES.get(action),
            dict(code=vcode),
        )

    set_vcode_info(request, mobile, vcode, action)
    return vcode


def sms_send(mobile, template_code, params):
    import uuid
    mobile = sanitize_mobile(mobile)
    __business_id = uuid.uuid1()

    if type(params) != str:
        params = json.dumps(params)

    # 检验发送成功与否并返回 True，失败的话抛一个错误
    from .setup_send import send_sms
    data = send_sms(__business_id, mobile, settings.SMS_SIGN_NAME, template_code, params)
    resp = json.loads(data.decode())
    if resp.get('Message') != 'OK':
        raise ValidationError(data)
    return resp


def vcode_required(action='', error=None, vcode_field='vcode', mobile_field='mobile'):
    def func(view_func):
        @wraps(view_func)
        def _wrapped_view(self, request, *args, **kwargs):
            vcode = request.data.get(vcode_field) or ''
            mobile = request.data.get(mobile_field) or ''
            result = validate_mobile_vcode(request, action, mobile, vcode)
            if not result:
                from django_base.base_utils.app_error.exceptions import AppError
                raise error or AppError(99999, '验证码错误', debug=False)
            return view_func(self, request, *args, **kwargs)

        return _wrapped_view

    return func



from django.conf import settings
from threading import currentThread
from django.utils.deprecation import MiddlewareMixin

_requests = {}


def get_request():
    return _requests[currentThread()]


class GlobalRequestMiddleware(MiddlewareMixin):
    """ 将 request 对象置于线程中可以供任意位置访问
    """

    @staticmethod
    def process_request(request):
        _requests[currentThread()] = request


class CustomExceptionMiddleware(MiddlewareMixin):
    """ 当捕获到 APIError 时自动包装响应一个错误的 HTTP 响应
    """

    @staticmethod
    def process_exception(request, exception):
        import traceback
        from django.http import JsonResponse
        from sys import stderr
        from django.conf import settings
        from django.core.exceptions import ValidationError
        # Retrieves the error message, response error message only.
        msg = exception.message if hasattr(exception, 'message') else str(exception)
        # Bypass the exception raised but still print to stderr
        if settings.DEBUG or type(exception) not in [AssertionError, ValidationError]:
            print(traceback.format_exc(), file=stderr)
        # Return a client-recognizable format.
        return JsonResponse(dict(
            ok=False,
            msg=msg
        ), status=400)


# class SingleSessionMiddleware(MiddlewareMixin):
#     def process_response(self, request, response):
#         from django.contrib.auth import logout
#         # from django.http import JsonResponse
#         # print(request.user.base_member.session_key)
#         # print(request.session.session_key)
#         if hasattr(request.user, 'base_member') and \
#                 request.user.base_member.session_key != \
#                 request.session.session_key and \
#                 request.get_full_path() != '/api/option/get_guide_image/' and \
#                 request.user.base_member.session_key:
#             import json
#             content_string = str(response.content, encoding="utf8")
#             dict_content = json.loads(content_string)
#             dict_content['force_logout'] = 1
#             response.content = bytes(json.dumps(dict_content), encoding="utf-8")
#             # print(content_string)
#             # print(response.content)
#             # print(response)
#             # print('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>')
#             return response
#         # if request.user.is_anonymous:
#         #     response['FORCE_LOGOUT'] = '1'
#         return response


class FullMediaUrlMiddleware(MiddlewareMixin):
    """ 在 MediaUrl 中自动拼接完整的 request 路径 """
    @staticmethod
    def process_request(request):
        from django.conf import settings
        from urllib.parse import urljoin
        settings.MEDIA_URL = \
            urljoin(request.get_raw_uri(), settings.MEDIA_URL)
        settings.STATIC_URL = \
            urljoin(request.get_raw_uri(), settings.STATIC_URL)
        # settings.ALIPAY_NOTIFY_URL = \
        #     urljoin(request.get_raw_uri(), settings.ALIPAY_NOTIFY_URL)


#
#
# class CookieCsrfMiddleware(MiddlewareMixin):
#     def process_request(self, request):
#         csrftoken = request.COOKIES.get('csrftoken')
#         if csrftoken:
#             request.META['HTTP_X_CSRFTOKEN'] = csrftoken


class ExplicitSessionMiddleware(MiddlewareMixin):
    """ 支持用户显式指定 SESSION_ID，以支持不用 COOKIE 时的情况 """

    def __init__(self, get_response=None):
        super().__init__(get_response)
        from importlib import import_module
        self.get_response = get_response
        engine = import_module(settings.SESSION_ENGINE)
        self.SessionStore = engine.SessionStore

    def process_request(self, request):
        # print('\n'.join(sorted(request.META.keys())))
        # session_id = request.META.get('HTTP_X_SESSIONID')
        # 实际的请求头到了 request.META 里面会变成 HTTP_ 前缀加上横杠变成下划线的转换
        session_id = request.META.get(
            ('HTTP_' + settings.CUSTOM_SESSION_HEADER).replace('-', '_')
        )
        # print(request.META.get('HTTP_X_CSRFTOKEN'))
        if session_id and self.SessionStore(session_id):
            request.COOKIES[settings.SESSION_COOKIE_NAME] = session_id

    def process_response(self, request, response):
        if request.session:
            session_id = getattr(request.session, 'session_key') or \
                         getattr(request.session, '_SessionBase__session_key')
            if session_id:
                response[settings.CUSTOM_SESSION_HEADER] = session_id
        return response


class MethodOverrideMiddleware(MiddlewareMixin):
    """
    用于支持客户端 PATCH 方法受限时，采取 POST 方法，但是添加一个 Header 来实现 PATCH 提交的功能
    Clone from https://pypi.org/project/django-method-override/
    To solve android volley request patch method no supported issue:
    https://stackoverflow.com/a/20221780/2544762
    """

    def process_request(self, request):
        if request.method != 'POST':
            return
        method = self._get_method_override(request)
        if method in settings.METHOD_OVERRIDE_ALLOWED_HTTP_METHODS:
            setattr(request, method, request.POST.copy())
            request.method = method

    def _get_method_override(self, request):
        method = (request.POST.get(settings.METHOD_OVERRIDE_PARAM_KEY) or
                  request.META.get(settings.METHOD_OVERRIDE_HTTP_HEADER))
        return method and method.upper()



from django.http import JsonResponse

import pickle

import requests
import os
import os.path
from uuid import uuid4

from django.conf import settings


class UserContext(object):
    """ 用于独立保存登录状态的 CookieJar 封装，可以持久化指定请求的 Cookie 以及部分 Header """
    SESSION_DIR = os.path.join(settings.MEDIA_ROOT, '.session')

    def __init__(self, context_id=''):
        self.context_id = context_id or str(uuid4())
        self.session_file_path = os.path.join(self.SESSION_DIR, self.context_id)
        if not os.path.isdir(os.path.dirname(self.session_file_path)):
            os.makedirs(os.path.dirname(self.session_file_path), exist_ok=True)

        if os.path.isfile(self.session_file_path):
            self.session = pickle.load(open(self.session_file_path, 'rb'))
        else:
            self.session = requests.Session()
        # 统一限定 UserAgent 免得有些站点闹别扭
        self.session.headers.update({
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_4) '
                          'AppleWebKit/603.1.30 (KHTML, like Gecko) Version/10.1 Safari/603.1.30'
        })

    def save(self):
        file = open(self.session_file_path, 'wb')
        pickle.dump(self.session, file)
        file.close()

    def delete(self):
        if os.path.isfile(self.session_file_path):
            os.remove(self.session_file_path)

    @classmethod
    def destroy(cls, context_id):
        """ 销毁一个 context_id 的缓存文件 """
        session_file = os.path.join(cls.SESSION_DIR, context_id)
        if os.path.isfile(session_file):
            os.remove(session_file)

    def print(self):
        [print('>>>>', k, '>>>>\n' + str(v), '\n<<<<\n') for k, v in self.__dict__.items()]


def response_success(msg='', *, data=None, silent=False):
    payload = dict(ok=True)
    if msg:
        payload['msg'] = msg
    if silent:
        payload['silent'] = True
    if data:
        payload['data'] = data
    return JsonResponse(payload, json_dumps_params=dict(ensure_ascii=False))


def response_fail(msg='', errcode=0, *, status=400, data=None, silent=False):
    payload = dict(ok=False)
    if msg:
        payload['msg'] = msg
    if silent:
        payload['silent'] = True
    if data:
        payload['data'] = data
    if errcode:
        payload['errcode'] = errcode
    return JsonResponse(payload, status=status, json_dumps_params=dict(ensure_ascii=False))



""" Utils for quick setup for your DRF project.
Usage:
------

If you are no familiar with django-rest-framework, see the below tutorials:

* https://www.django-rest-framework.org/tutorial/1-serialization/#creating-a-serializer-class
* https://www.django-rest-framework.org/tutorial/6-viewsets-and-routers/#tutorial-6-viewsets-routers

So, with this module, you don't need to generate the ModalSerializer classes and ViewSet.

Just call `auto_declare_serializers` and `auto_declare_viewsets` in your code, everything is done.

Example:

In your `serializers.py`:

```
from rest_framework import serializers
from scaffold.restframework.utils import auto_declare_serializers, extend_class

# Import your models module where Model classes were defined.
from . import models as m

# Auto declare serializer classes from models.
auto_declare_serializers(m, locals())

# Optional: Type hinting for Serializer classes generated.
MemberSerializer: serializers.Serializer


#  >>> Extending serializer classes as needed.

@extend_class(MemberSerializer)
class Extender: # Use a any class name as you like except for overriding MemberSerializer itself.
    avatar_url = serializers.ReadOnlyField(source='avatar.url')
```

In your `views.py`:

```
from scaffold.restframework.utils import auto_declare_viewsets

# Automatically generated ViewSet classes from serializers.
auto_declare_viewsets(s, locals())
```

"""
import inspect
import logging
import re
from types import ModuleType
from typing import Union, List

from django.db import models
from rest_framework import viewsets, serializers
from rest_framework.routers import DefaultRouter
from rest_framework.viewsets import ViewSet

logger = logging.getLogger(__name__)


def auto_declare_serializers(models_module, context):
    """ Automatically declares classes from serializers
    :param models_module: Passes the module to search model classes.
    :param context: Context module to export classes, should passes locals().
    """
    for model in models_module.__dict__.values():
        if not inspect.isclass(model) \
                or not issubclass(model, models.Model) \
                or model._meta.abstract \
                or not model._meta.managed \
                or model._meta.app_label != context['__package__']:
            continue
        serializer_name = model.__name__ + 'Serializer'
        # Do not override
        if serializer_name in context:
            continue
        # Derive subclass of serializers.Serializer
        serializer = type(
            serializer_name,
            (serializers.ModelSerializer,),
            dict(
                Meta=type('Meta', (object,), dict(model=model, fields='__all__')),
            )
        )
        logger.debug(f'>>> Automatically declared <class \'{serializer.__name__}\'>')
        context[serializer.__name__] = serializer


def auto_declare_viewsets(serializers_module, context):
    """ Automatically declares classes from serializers
    :param serializers_module: Passes the module to search serializer classes.
    :param context: Context module to export classes, should passes locals().
    """
    for serializer in serializers_module.__dict__.values():
        if not inspect.isclass(serializer) \
                or not issubclass(serializer, serializers.ModelSerializer):
            continue
        model = getattr(serializer, 'Meta').model
        # Skip inconsist app, specifically when you import a Model from another app.
        if model._meta.app_label != context['__package__']:
            continue
        viewset_name = model.__name__ + 'ViewSet'
        # Do not override
        if viewset_name in context:
            continue
        # Dynamic declare the subclass
        view_set = type(
            viewset_name,
            (viewsets.ModelViewSet,),
            dict(
                queryset=model.objects.all(),
                serializer_class=serializer,
                filter_fields='__all__',
                ordering=['-pk']
            )
        )
        logger.debug(f'>>> Automatically declared <class \'{view_set.__name__}\'>')
        context[view_set.__name__] = view_set


def extend_class(klass):
    """ Dynamically extends a class with another class definition
    :param klass: The class to by modified.
    """

    def klass_decorator(cls):
        """ The inner class which was inline defined and properties should be used for injection.
        :param cls: The class to by modified.
        """
        for k, v in cls.__dict__.items():
            if k in ('__module__', '__dict__'):
                continue
            if v is None and hasattr(klass, k):
                delattr(klass, k)
            else:
                setattr(klass, k, v)

    return klass_decorator


def auto_collect_urls(modules: Union[ModuleType, List[ModuleType]]):
    """ Automatically collect ViewSet classes to urls list for `urls.py`
    :param modules: passing module(s)
    :return:
    """
    router = DefaultRouter()

    routers = []

    # Fall-backs for single-module usage
    if type(modules) == ModuleType:
        modules = [modules]

    for module in modules:
        for key, item in module.__dict__.items():
            # Catches name ends with ViewSet.
            # 检验 queryset 存在这个条件主要是为了排除 GenericViewSet 等非自定义 ViewSet 类被捕捉到
            if key.endswith('ViewSet') and getattr(item, 'queryset', None) is not None:
                # Replacing ViewSet from uppercase to underscored naming as Rest resource name.
                name = key.replace('ViewSet', '')
                name = re.sub(r'([A-Z])', '_\\1', name)[1:].lower()
                if name:
                    # 如果name设置成功，则添加到routers数组中
                    routers.append((name, item))

    for name, item in routers:
        # print(name, item)
        # 进行视图集路由注册 register(prefix(视图集的路由前缀),viewset(视图集),base_name(路由名称的前缀))
        router.register(name, item)

    # Return urls
    return router.urls



from django.contrib.auth.decorators import login_required
from django.contrib.contenttypes.models import ContentType
from django.db.models.base import ModelBase
from django.utils.decorators import method_decorator
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.response import Response

from scaffold import utils as u

from . import models as m
from . import serializers as s


class OptionViewSet(viewsets.GenericViewSet):
    queryset = m.Option.objects.all()
    serializer_class = s.OptionSerializer
    filter_fields = '__all__'

    @action(methods=['GET'], detail=False)
    def get_all(self, request):
        return Response(m.Option.get_all())

    @action(methods=['GET'], detail=False)
    def get(self, request):
        return Response(m.Option.get(request.query_params.get('key')))

    @action(methods=['POST'], detail=False)
    def set(self, request):
        m.Option.set(request.data.get('key'), request.data.get('value'))
        return u.http.response_success('设置成功', silent=True)

    @action(methods=['GET'], detail=False)
    def choices(self, request):
        """ 返回所有的常量选项 """
        from django.apps import apps
        result = dict()

        def inspect_class(cls):
            if type(cls) != ModelBase or cls.__name__ in result or not hasattr(cls, '_meta'):
                return
            section = dict()
            if not cls._meta.abstract:
                section['content_type_id'] = ContentType.objects.get_for_model(cls).id
            if hasattr(cls, '__doc__'):
                section['help_text'] = cls.__doc__
            for attr_name, attr in cls.__dict__.items():
                # print(cls.__name__, attr_name, attr)
                if attr_name.endswith('_CHOICES'):
                    section[attr_name.replace('_CHOICES', '').lower()] = dict(attr)
            result[cls.__name__] = section
            # 深入解析所有的 abstract base class
            for super_class in cls.__bases__:
                inspect_class(super_class)

        for app in apps.all_models.values():
            for cls in app.values():
                inspect_class(cls)
        return Response(result)


class UserOptionViewSet(viewsets.GenericViewSet):
    queryset = m.UserOption.objects.all()
    serializer_class = s.UserOptionSerializer
    filter_fields = '__all__'

    @action(methods=['GET'], detail=False)
    @method_decorator(login_required)
    def get_all(self, request):
        user = request.user
        return Response(m.UserOption.get_all(user))

    @action(methods=['GET'], detail=False)
    @method_decorator(login_required)
    def get(self, request):
        user = request.user
        return Response(m.UserOption.get(user, request.query_params.get('key')))

    @action(methods=['POST'], detail=False)
    @method_decorator(login_required)
    def set(self, request):
        user = request.user
        m.UserOption.set(user, request.data.get('key'), request.data.get('value'))
        return u.http.response_success('设置成功', silent=True)


class VersionViewSet(viewsets.ModelViewSet):
    queryset = m.Version.objects.all()
    serializer_class = s.VersionSerializer
    filter_fields = '__all__'
    ordering = ['-pk']



from django.conf import settings
from django.db import models


# class Menu(HierarchicalModel, models.Model):
#     seq = models.IntegerField(
#         verbose_name='序号',
#         blank=True,
#         default=0,
#     )
#
#     project = models.CharField(
#         verbose_name='项目名称',
#         max_length=50,
#         blank=True,
#         default='',
#         help_text='如果有多个后台模块，通过这个字段区分',
#     )
#
#     groups = models.ManyToManyField(
#         verbose_name='组',
#         to=Group,
#         related_name='menus',
#     )
#
#     name = models.CharField(
#         verbose_name='菜单名称',
#         max_length=100,
#         help_text='建议与路由名称同步',
#     )
#
#     title = models.CharField(
#         verbose_name='菜单标题',
#         max_length=150,
#     )
#
#     class Meta:
#         verbose_name = '菜单'
#         verbose_name_plural = '菜单'
#         db_table = 'base_menu'
#         ordering = ['seq']
#
#     def __str__(self):
#         return '%s: %s' % (self.name, self.title)


class Option(models.Model):
    """ 全局选项
    可以通过设置的 DEFAULT_OPTIONS 来指定默认的选项值
    ```
    DEFAULT_OPTIONS: dict(
        auto_login=True,
        sound=True,
        vibration=False,
    )
    ```
    """

    key = models.CharField(
        verbose_name='选项关键字', max_length=128, unique=True)

    value = models.TextField(
        verbose_name='选项值', blank=True, default='')

    class Meta:
        verbose_name = '系统选项'
        verbose_name_plural = '系统选项'
        db_table = 'base_config_option'

    def __str__(self):
        return '{} = {}'.format(self.key, self.value)

    @classmethod
    def get(cls, key):
        """ 获取选项值
        :param key: 选项的关键字
        :return: 匹配到的选项值，如果没有此选项，返回 None
        """
        opt = cls.objects.filter(key=key).first()
        return opt and opt.value

    @classmethod
    def unset(cls, key):
        """ 删除选项值
        :param key: 选项的关键字
        :return: 没有返回值
        """
        cls.objects.filter(key=key).delete()

    @classmethod
    def set(cls, key, val):
        """ 设置选项值
        :param key: 选项的关键字
        :param val: 需要设置的目标值
        :return: 没有返回值
        """
        opt, created = cls.objects.get_or_create(key=key)
        opt.value = val
        opt.save()

    @staticmethod
    def get_all():
        return dict([(item.key, item.value) for item in Option.objects.all()])


class UserOption(models.Model):
    """ 用户选项
    类似于全局选项，但是选项配置是根据用户独立配置的
    可以通过设置的 DEFAULT_OPTIONS 来指定默认的选项值
    ```
    DEFAULT_USER_OPTIONS: dict(
        auto_login=True,
        sound=True,
        vibration=False,
    )
    ```
    """

    user = models.ForeignKey(
        verbose_name='用户',
        to='auth.User',
        related_name='options',
        on_delete=models.CASCADE,
    )

    key = models.CharField(
        verbose_name='选项名', max_length=128)

    value = models.TextField(
        verbose_name='选项值', blank=True, default='')

    class Meta:
        verbose_name = '用户选项'
        verbose_name_plural = '用户选项'
        unique_together = [('user', 'key')]
        db_table = 'base_config_user_option'

    def __str__(self):
        return '{}: {} = {}'.format(self.user, self.key, self.value)

    @staticmethod
    def get_all(user):
        return dict([(item.key, item.value) for item in user.options.all()])

    @classmethod
    def get(cls, user, key):
        """ 获取选项值
        :param key: 选项的关键字
        :param user: 选项对应的用户
        :return: 匹配到的选项值，如果没有此选项，返回 None
        """
        opt = cls.objects.filter(user=user, key=key).first()
        return opt and opt.value

    @classmethod
    def set(cls, user, key, val):
        """ 设置选项值
        :param user: 用户
        :param key: 选项的关键字
        :param val: 需要设置的目标值
        :return: 没有返回值
        """
        opt, created = cls.objects.get_or_create(key=key, user=user)
        opt.value = val
        opt.save()


class Version(models.Model):
    version = models.CharField(
        verbose_name='版本号',
        max_length=20,
    )

    alias = models.CharField(
        verbose_name='版本别名',
        max_length=100,
        blank=True,
        default='',
    )

    platform = models.CharField(
        verbose_name='平台',
        max_length=20,
        blank=True,
        default='',
    )

    date_created = models.DateTimeField(
        verbose_name='创建时间',
        auto_now_add=True,
    )

    date_updated = models.DateTimeField(
        verbose_name='更新时间',
        auto_now=True,
    )

    is_active = models.BooleanField(
        verbose_name='是否可用',
        default=True,
    )

    is_master = models.BooleanField(
        verbose_name='是否主要版本',
        default=True,
        help_text='建议实践：如果是主要版本，所有低版本的客户端必须要升级到此版本才能使用，'
                  '否则为建议版本，只作提醒，不强制升级。',
    )

    description = models.TextField(
        verbose_name='版本描述',
        blank=True,
        default='',
    )

    link = models.CharField(
        verbose_name='下载链接',
        max_length=255,
        blank=True,
        default='',
    )

    attachment = models.OneToOneField(
        verbose_name='安装包附件',
        to='media.Attachment',
        related_name='version',
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )

    class Meta:
        verbose_name = '版本'
        verbose_name_plural = '版本'
        db_table = 'base_version'
        unique_together = ''



""" Extended rest_framework permissions classes """
import re
from logging import getLogger

from django.conf import settings
from rest_framework.permissions import BasePermission, SAFE_METHODS

logger = getLogger(__name__)


class ActionBasedPermission(BasePermission):
    """ ViewSet action level Permission integrator
    https://stackoverflow.com/a/47528633/2544762
    Grant or deny access to a view, based on a mapping in view.action_permissions

    Example:

        class MyModelViewSet(ReadOnlyModelViewSet):
            serializer_class = MyModelSerializer
            queryset = MyModel.objects.all()

            permission_classes = [ActionBasedPermission]
            action_permissions = {
                'list,retrieve': IsMember,
                'destroy': IsAdminUser,
                'create,update,partial_update':
                    HasPermissions.build('basic_info_admin'),
            }
    """

    def has_permission(self, request, view):
        for actions, cls in getattr(view, 'action_permissions', {}).items():
            # 支持 actions 是字符串或者数组，如果是字符串先转为数组
            if isinstance(actions, str):
                actions = re.split(r'\s+|[\.\|,]', actions)
            # 判断是当前 view 的动作是否在清单中
            if view.action in actions:
                return cls().has_permission(request, view)
        return True

    def has_object_permission(self, request, view, obj):
        for actions, cls in getattr(view, 'action_permissions', {}).items():
            # 支持 actions 是字符串或者数组，如果是字符串先转为数组
            if isinstance(actions, str):
                actions = re.split(r'\s+|[\.\|,]', actions)
            # 判断是当前 view 的动作是否在清单中
            if view.action in actions:
                return cls().has_object_permission(request, view, obj)
        return True


class IsAuthor(BasePermission):
    """ has permission if the current user is the author of the object """

    def has_permission(self, request, view):
        return request.user.is_authenticated

    def has_object_permission(self, request, view, obj):
        return obj.author == request.user


# TODO: Planning to deprecate in 1.0
class IsAdminOrIsSelf(BasePermission):
    """ has permission if the current user is the owner of the object or admin user """

    def has_object_permission(self, request, view, obj):
        logger.warning(
            'Do not use IsAdminOrIsSelf permission, '
            'use bitwise OR to compose IsAuthor and IsAdminUser instead.')
        return obj.author == request.user or \
               request.user.is_staff or \
               request.user.is_superuser


class IsAdminOrReadOnly(BasePermission):
    """ has permission only if the current user is admin """

    def has_permission(self, request, view):
        return request.method in SAFE_METHODS or \
               request.user and (request.user.is_staff or request.user.is_superuser)


class HasPermissions(BasePermission):
    """ check if user has a specific django permission

    Example:

        class MyModelViewSet(ReadOnlyModelViewSet):
            serializer_class = MyModelSerializer
            queryset = MyModel.objects.all()

            permission_classes = (
                HasPermissions.build('core.sales_manager', 'core.user_manager'),
            )

    """
    # if no app name specified, e.g. 'my_permission' rather than 'myapp.my_permissions',
    # a default app label will be added, you can set another app in setting as needed.
    app_label = getattr(settings, 'DEFAULT_PERMISSION_APP', 'core')
    perms = ()

    @classmethod
    def build(cls, *args):
        """ Derive a meta subclass of current by replacing the perms property. """
        return type(
            'HasPermissions',
            (cls,),
            dict(perms=tuple(p if '.' in p else f'{cls.app_label}.{p}' for p in args))
        )

    def has_permission(self, request, view):
        if not request.user or not request.user.is_authenticated:
            return False
        return request.user.is_superuser or any(request.user.has_perm(perm) for perm in self.perms)

    def has_object_permission(self, request, view, obj):
        return self.has_permission(request, view)
